#ifndef CAFFE_BASE_WEIGHTED_L2_SIMILARITY_LAYER_HPP_
#define CAFFE_BASE_WEIGHTED_L2_SIMILARITY_LAYER_HPP_

#include <vector>

#include "caffe/blob.hpp"
#include "caffe/layer.hpp"
#include "caffe/proto/caffe.pb.h"
#include "caffe/util/im2col.hpp"

namespace caffe {

/**
 * @brief Abstract base class that factors out the BLAS code common to
 *        WeightedL2ConvolutionLayer and DeconvolutionLayer.
 */
template <typename Dtype>
class BaseWeightedL2ConvolutionLayer : public Layer<Dtype> {
 public:
  explicit BaseWeightedL2ConvolutionLayer(const LayerParameter& param)
      : Layer<Dtype>(param) {}
  virtual void LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);
  virtual void Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);

  virtual inline int MinBottomBlobs() const { return 1; }
  virtual inline int MinTopBlobs() const { return 1; }
  virtual inline bool EqualNumBottomTopBlobs() const { return true; }

 protected:
  // Helper functions that abstract away the column buffer and gemm arguments.
  // The last argument in forward_cpu_gemm is so that we can skip the im2col if
  // we just called scale_cpu_gemm with the same input.
  void set_blobs(int M, int N, int K);
  void set_weights_cpu(const Dtype* means,const Dtype* sclales);

  void forward_cpu_gemm(const Dtype* input, const Dtype* means, const Dtype* scales,
      Dtype* output_diff, bool skip_im2col = false);
  void forward_cpu_bias(Dtype* output, const Dtype* bias);
  
//   void backward_cpu_gemm(const Dtype* output_diff, const Dtype* means, const Dtype* scales,
//       Dtype* input_diff);
//   void mean_scale_cpu_gemm(const Dtype* input, const Dtype* output_diff, const Dtype*
//       means, Dtype* mean_diff, const Dtype* scales, Dtype* scale_diff);

    void backward_cpu(const Dtype* input, const Dtype* output_diff, const Dtype* means,
		      const Dtype* scales, Dtype* input_diff, Dtype* means_diff, Dtype* scales_diff);
  
  void backward_cpu_bias(Dtype* bias, const Dtype* input);

#ifndef CPU_ONLY
  void forward_gpu_gemm(const Dtype* input, const Dtype* means, const Dtype* scales,
      Dtype* output_diff, bool skip_im2col = false);
  void forward_gpu_bias(Dtype* output, const Dtype* bias);
  void backward_gpu(const Dtype* input, const Dtype* output_diff, const Dtype* means,
		      const Dtype* scales, Dtype* input_diff, Dtype* means_diff, Dtype* scales_diff);
  //void scale_gpu_gemm(const Dtype* col_input, const Dtype* output, Dtype*
  //    scales);
  void backward_gpu_bias(Dtype* bias, const Dtype* input);
  void set_weights_gpu(const Dtype* means,const Dtype* sclales);
#endif

  /// @brief The spatial dimensions of the input.
  inline int input_shape(int i) {
    return (*bottom_shape_)[channel_axis_ + i];
  }
  // reverse_dimensions should return true iff we are implementing deconv, so
  // that conv helpers know which dimensions are which.
  virtual bool reverse_dimensions() = 0;
  // Compute height_out_ and width_out_ from other parameters.
  virtual void compute_output_shape() = 0;

  /// @brief The spatial dimensions of a filter kernel.
  Blob<int> kernel_shape_;
  /// @brief The spatial dimensions of the stride.
  Blob<int> stride_;
  /// @brief The spatial dimensions of the padding.
  Blob<int> pad_;
  /// @brief The spatial dimensions of the dilation.
  Blob<int> dilation_;
  /// @brief The spatial dimensions of the convolution input.
  Blob<int> conv_input_shape_;
  /// @brief The spatial dimensions of the col_buffer.
  vector<int> col_buffer_shape_;
  /// @brief The spatial dimensions of the output.
  vector<int> output_shape_;
  const vector<int>* bottom_shape_;

  int num_spatial_axes_;
  int bottom_dim_;
  int top_dim_;

  int channel_axis_;
  int num_;
  int channels_;
  int group_;
  int out_spatial_dim_;
  int weight_offset_;
  int num_output_;
  bool bias_term_;
  bool is_1x1_;
  bool force_nd_im2col_;

 private:
  // wrap im2col/col2im so we don't have to remember the (long) argument lists
  inline void conv_im2col_cpu(const Dtype* data, Dtype* col_buff) {
    if (!force_nd_im2col_ && num_spatial_axes_ == 2) {
      im2col_cpu(data, conv_in_channels_,
          conv_input_shape_.cpu_data()[1], conv_input_shape_.cpu_data()[2],
          kernel_shape_.cpu_data()[0], kernel_shape_.cpu_data()[1],
          pad_.cpu_data()[0], pad_.cpu_data()[1],
          stride_.cpu_data()[0], stride_.cpu_data()[1],
          dilation_.cpu_data()[0], dilation_.cpu_data()[1], col_buff);
    } else {
      im2col_nd_cpu(data, num_spatial_axes_, conv_input_shape_.cpu_data(),
          col_buffer_shape_.data(), kernel_shape_.cpu_data(),
          pad_.cpu_data(), stride_.cpu_data(), dilation_.cpu_data(), col_buff);
    }
  }
  inline void conv_col2im_cpu(const Dtype* col_buff, Dtype* data) {
    if (!force_nd_im2col_ && num_spatial_axes_ == 2) {
      col2im_cpu(col_buff, conv_in_channels_,
          conv_input_shape_.cpu_data()[1], conv_input_shape_.cpu_data()[2],
          kernel_shape_.cpu_data()[0], kernel_shape_.cpu_data()[1],
          pad_.cpu_data()[0], pad_.cpu_data()[1],
          stride_.cpu_data()[0], stride_.cpu_data()[1],
          dilation_.cpu_data()[0], dilation_.cpu_data()[1], data);
    } else {
      col2im_nd_cpu(col_buff, num_spatial_axes_, conv_input_shape_.cpu_data(),
          col_buffer_shape_.data(), kernel_shape_.cpu_data(),
          pad_.cpu_data(), stride_.cpu_data(), dilation_.cpu_data(), data);
    }
  }
#ifndef CPU_ONLY
  inline void conv_im2col_gpu(const Dtype* data, Dtype* col_buff) {
    if (!force_nd_im2col_ && num_spatial_axes_ == 2) {
      im2col_gpu(data, conv_in_channels_,
          conv_input_shape_.cpu_data()[1], conv_input_shape_.cpu_data()[2],
          kernel_shape_.cpu_data()[0], kernel_shape_.cpu_data()[1],
          pad_.cpu_data()[0], pad_.cpu_data()[1],
          stride_.cpu_data()[0], stride_.cpu_data()[1],
          dilation_.cpu_data()[0], dilation_.cpu_data()[1], col_buff);
    } else {
      im2col_nd_gpu(data, num_spatial_axes_, num_kernels_im2col_,
          conv_input_shape_.gpu_data(), col_buffer_.gpu_shape(),
          kernel_shape_.gpu_data(), pad_.gpu_data(),
          stride_.gpu_data(), dilation_.gpu_data(), col_buff);
    }
  }
  inline void conv_col2im_gpu(const Dtype* col_buff, Dtype* data) {
    if (!force_nd_im2col_ && num_spatial_axes_ == 2) {
      col2im_gpu(col_buff, conv_in_channels_,
          conv_input_shape_.cpu_data()[1], conv_input_shape_.cpu_data()[2],
          kernel_shape_.cpu_data()[0], kernel_shape_.cpu_data()[1],
          pad_.cpu_data()[0], pad_.cpu_data()[1],
          stride_.cpu_data()[0], stride_.cpu_data()[1],
          dilation_.cpu_data()[0], dilation_.cpu_data()[1], data);
    } else {
      col2im_nd_gpu(col_buff, num_spatial_axes_, num_kernels_col2im_,
          conv_input_shape_.gpu_data(), col_buffer_.gpu_shape(),
          kernel_shape_.gpu_data(), pad_.gpu_data(), stride_.gpu_data(),
          dilation_.gpu_data(), data);
    }
  }
#endif

  int num_kernels_im2col_;
  int num_kernels_col2im_;
  int conv_out_channels_;
  int conv_in_channels_;
  int conv_out_spatial_dim_;
  int kernel_dim_;
  int col_offset_;
  int output_offset_;
  WeightedL2ConvolutionParameter_ScaleParameterization	scale_parameterization_;
  Dtype alpha_;

  Blob<Dtype> rep_mean_buffer_;
  Blob<Dtype> col_mean_diff_sqr_buffer_;
  Blob<Dtype> col_mean_diff_sqr_deriv_buffer_;
  Blob<Dtype> col_buffer_;
  Blob<Dtype> col_buffer_diff_;
  Blob<Dtype> col_buffer_diff_total_;
  Blob<Dtype> bias_multiplier_;

  Blob<Dtype> processed_scales;
  Blob<Dtype> s_m_dot;
  Blob<Dtype> input_power_2;
  Blob<Dtype> ones_n;
  Blob<Dtype> foward_MSM;
  Blob<Dtype> backward_mean_power_2;
  Blob<Dtype> backward_sign_scale;

  Blob<Dtype> forward_temp_1;
  Blob<Dtype> backward_temp_1;
  Blob<Dtype> backward_temp_2;
  Blob<Dtype> backward_temp_3;
  Blob<Dtype> backward_temp_4;
  Blob<Dtype> backward_temp_5;
  Blob<Dtype> backward_temp_6;



  bool	      natural_gradient_;
  bool	      negate_distance_;
  bool	      zero_mean_;
  int	      rev_fan_in_;
};

}  // namespace caffe

#endif  // CAFFE_BASE_CONVOLUTION_LAYER_HPP_
