#include <vector>

#include "caffe/filler.hpp"
#include "caffe/layer.hpp"
#include "caffe/layers/adaptive_exp_layer.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {

template <typename Dtype>
void AdaptiveExpLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  NeuronLayer<Dtype>::LayerSetUp(bottom, top);
  coef_  = this->layer_param_.adaptive_exp_param().coef();

  this->blobs_.resize(1);
  this->blobs_[0].reset(new Blob<Dtype>(1,1,1,1));
  this->param_propagate_down_.resize(this->blobs_.size(), true);
  this->param_propagate_down_[0] = true;

  shared_ptr<Filler<Dtype> > scale_filler(GetFiller<Dtype>(
		this->layer_param_.adaptive_exp_param().scale_filler()));
  scale_filler->Fill(this->blobs_[0].get());
}

template <typename Dtype>
void AdaptiveExpLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
    const vector<Blob<Dtype>*>& top) {
  const int count = bottom[0]->count();
  const Dtype* bottom_data = bottom[0]->cpu_data();
  Dtype* top_data = top[0]->mutable_cpu_data();
  const Dtype* scale;

  scale = this->blobs_[0]->cpu_data();
  caffe_cpu_scale(count, coef_ * (*scale) * (*scale), bottom_data, top_data);
  caffe_exp(count, top_data, top_data);

}

template <typename Dtype>
void AdaptiveExpLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {
  static unsigned int scount = 0;
  const int count = bottom[0]->count();
  const Dtype* top_data = top[0]->cpu_data();
  const Dtype* top_diff = top[0]->cpu_diff();
  Dtype* scale = this->blobs_[0]->mutable_cpu_data();
  Dtype* scale_diff = this->blobs_[0]->mutable_cpu_diff();
  Dtype* bottom_diff = bottom[0]->mutable_cpu_diff();
  const Dtype* bottom_data = bottom[0]->cpu_data();

  if (this->param_propagate_down_[0])
  {
	caffe_mul(count, top_data, top_diff, bottom_diff);
	caffe_mul(count, bottom_diff, bottom_data, bottom_diff);
	caffe_scal(count, Dtype(2.0) * coef_ * (*scale), bottom_diff);
	*scale_diff = caffe_cpu_sum(count, bottom_diff);
  }

  if (propagate_down[0])
  {
 	caffe_mul(count, top_data, top_diff, bottom_diff);
 	caffe_scal(count, coef_ * (*scale) * (*scale), bottom_diff);
  }

//   if (scount % 1 == 0)
//   {
//     std::cout << std::endl;
//     std::cout << "AdaptiveExpLayer::scale=" << *scale<< std::endl;
//     std::cout << "AdaptiveExpLayer::scale_diff=" << scale_diff[0]<< std::endl;
//     std::cout << "AdaptiveExpLayer::bottom_diff.norm()=" << caffe_norm2(bottom[0]->count(),bottom_diff)<< std::endl;
//   }
  scount++;
}

#ifdef CPU_ONLY
STUB_GPU(AdaptiveExpLayer);
#endif

INSTANTIATE_CLASS(AdaptiveExpLayer);
REGISTER_LAYER_CLASS(AdaptiveExp);

}  // namespace caffe
