// Authors:   Keivan Nalaie 
//            Kamaledin Ghiasi-Shirazi

#include <algorithm>
#include <vector>

#include "caffe/filler.hpp"
#include "caffe/layers/base_weighted_l2_convolution_layer.hpp"
#include "caffe/util/im2col.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::set_blobs(int M,int N, int K){
	vector<int>shapes;

//  M: conv_out_channels_
//	N: conv_out_spatial_dim_
//	K: kernel_dim_
	shapes.clear();
	shapes.push_back(M);
	shapes.push_back(K);
	shapes.push_back(1);
	processed_scales.Reshape(shapes); // setup a blob for scale parameter if it is needed to process it
	s_m_dot.Reshape(shapes);
	backward_mean_power_2.Reshape(shapes); // setup a blob for mean^2
	backward_sign_scale.Reshape(shapes);
	shapes.clear();
	shapes.push_back(N);
	shapes.push_back(K);
	shapes.push_back(group_);
	input_power_2.Reshape(shapes); // setup a blob for input^2
	shapes.clear();
	shapes.push_back(N);
	shapes.push_back(1);
	shapes.push_back(1);
	ones_n.Reshape(shapes);
	if (Caffe::mode() == Caffe::GPU) {
#ifndef CPU_ONLY
		caffe_gpu_set<Dtype>(N, Dtype(1), ones_n.mutable_gpu_data());
#else
		NO_GPU;
#endif
	} else {
		caffe_set<Dtype>(N, Dtype(1), ones_n.mutable_cpu_data());
	}
	shapes.clear();
	shapes.push_back(M/group_);
	shapes.push_back(M/group_);
	shapes.push_back(1);
	forward_temp_1.Reshape(shapes);
	shapes.clear();
	shapes.push_back(M);
	shapes.push_back(1);
	shapes.push_back(1);
	foward_MSM.Reshape(shapes);       // setup a blob for Hadamard product of mean^2 and scale
	backward_temp_1.Reshape(shapes);
	shapes.clear();
	shapes.push_back(M/group_);
	shapes.push_back(M/group_);
	shapes.push_back(1);
	backward_temp_2.Reshape(shapes);
	shapes.clear();
	shapes.push_back(K);
	shapes.push_back(N);
	shapes.push_back(1);
	backward_temp_3.Reshape(shapes);
	shapes.clear();
	shapes.push_back(M/group_);
	shapes.push_back(K);
	shapes.push_back(1);
	backward_temp_4.Reshape(shapes);
	backward_temp_5.Reshape(shapes);
	backward_temp_6.Reshape(shapes);



}
template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top) {
	// Configure the kernel size, padding, stride, and inputs.
	WeightedL2ConvolutionParameter conv_param = this->layer_param_.weighted_l2_convolution_param();
	negate_distance_  = conv_param.negate_distance();
	zero_mean_  = conv_param.zero_mean();
	natural_gradient_ = conv_param.natural_gradient();
        alpha_ = (Dtype) (conv_param.euclid_dotprod_mixing_coef_percent() / 100.0);

	scale_parameterization_ = conv_param.scale_parameterization();
	force_nd_im2col_ = conv_param.force_nd_im2col();
	channel_axis_ = bottom[0]->CanonicalAxisIndex(conv_param.axis());
	const int first_spatial_axis = channel_axis_ + 1;
	const int num_axes = bottom[0]->num_axes();
	num_spatial_axes_ = num_axes - first_spatial_axis;
	CHECK_GE(num_spatial_axes_, 0);
	vector<int> bottom_dim_blob_shape(1, num_spatial_axes_ + 1);
	vector<int> spatial_dim_blob_shape(1, std::max(num_spatial_axes_, 1));
	// Setup filter kernel dimensions (kernel_shape_).
	kernel_shape_.Reshape(spatial_dim_blob_shape);
	int* kernel_shape_data = kernel_shape_.mutable_cpu_data();
	if (conv_param.has_kernel_h() || conv_param.has_kernel_w()) {
		CHECK_EQ(num_spatial_axes_, 2)
        																												<< "kernel_h & kernel_w can only be used for 2D convolution.";
		CHECK_EQ(0, conv_param.kernel_size_size())
		<< "Either kernel_size or kernel_h/w should be specified; not both.";
		kernel_shape_data[0] = conv_param.kernel_h();
		kernel_shape_data[1] = conv_param.kernel_w();
	} else {
		const int num_kernel_dims = conv_param.kernel_size_size();
		CHECK(num_kernel_dims == 1 || num_kernel_dims == num_spatial_axes_)
		<< "kernel_size must be specified once, or once per spatial dimension "
		<< "(kernel_size specified " << num_kernel_dims << " times; "
		<< num_spatial_axes_ << " spatial dims).";
		for (int i = 0; i < num_spatial_axes_; ++i) {
			kernel_shape_data[i] =
					conv_param.kernel_size((num_kernel_dims == 1) ? 0 : i);
		}
	}
	for (int i = 0; i < num_spatial_axes_; ++i) {
		CHECK_GT(kernel_shape_data[i], 0) << "Filter dimensions must be nonzero.";
	}
	// Setup stride dimensions (stride_).
	stride_.Reshape(spatial_dim_blob_shape);
	int* stride_data = stride_.mutable_cpu_data();
	if (conv_param.has_stride_h() || conv_param.has_stride_w()) {
		CHECK_EQ(num_spatial_axes_, 2)
												<< "stride_h & stride_w can only be used for 2D convolution.";
		CHECK_EQ(0, conv_param.stride_size())
		<< "Either stride or stride_h/w should be specified; not both.";
		stride_data[0] = conv_param.stride_h();
		stride_data[1] = conv_param.stride_w();
	} else {
		const int num_stride_dims = conv_param.stride_size();
		CHECK(num_stride_dims == 0 || num_stride_dims == 1 ||
				num_stride_dims == num_spatial_axes_)
		<< "stride must be specified once, or once per spatial dimension "
		<< "(stride specified " << num_stride_dims << " times; "
		<< num_spatial_axes_ << " spatial dims).";
		const int kDefaultStride = 1;
		for (int i = 0; i < num_spatial_axes_; ++i) {
			stride_data[i] = (num_stride_dims == 0) ? kDefaultStride :
					conv_param.stride((num_stride_dims == 1) ? 0 : i);
			CHECK_GT(stride_data[i], 0) << "Stride dimensions must be nonzero.";
		}
	}
	// Setup pad dimensions (pad_).
	pad_.Reshape(spatial_dim_blob_shape);
	int* pad_data = pad_.mutable_cpu_data();
	if (conv_param.has_pad_h() || conv_param.has_pad_w()) {
		CHECK_EQ(num_spatial_axes_, 2)
												<< "pad_h & pad_w can only be used for 2D convolution.";
		CHECK_EQ(0, conv_param.pad_size())
		<< "Either pad or pad_h/w should be specified; not both.";
		pad_data[0] = conv_param.pad_h();
		pad_data[1] = conv_param.pad_w();
	} else {
		const int num_pad_dims = conv_param.pad_size();
		CHECK(num_pad_dims == 0 || num_pad_dims == 1 ||
				num_pad_dims == num_spatial_axes_)
		<< "pad must be specified once, or once per spatial dimension "
		<< "(pad specified " << num_pad_dims << " times; "
		<< num_spatial_axes_ << " spatial dims).";
		const int kDefaultPad = 0;
		for (int i = 0; i < num_spatial_axes_; ++i) {
			pad_data[i] = (num_pad_dims == 0) ? kDefaultPad :
					conv_param.pad((num_pad_dims == 1) ? 0 : i);
		}
	}
	// Setup dilation dimensions (dilation_).
	dilation_.Reshape(spatial_dim_blob_shape);
	int* dilation_data = dilation_.mutable_cpu_data();
	const int num_dilation_dims = conv_param.dilation_size();
	CHECK(num_dilation_dims == 0 || num_dilation_dims == 1 ||
			num_dilation_dims == num_spatial_axes_)
	<< "dilation must be specified once, or once per spatial dimension "
	<< "(dilation specified " << num_dilation_dims << " times; "
	<< num_spatial_axes_ << " spatial dims).";
	const int kDefaultDilation = 1;
	for (int i = 0; i < num_spatial_axes_; ++i) {
		dilation_data[i] = (num_dilation_dims == 0) ? kDefaultDilation :
				conv_param.dilation((num_dilation_dims == 1) ? 0 : i);
	}
	// Special case: im2col is the identity for 1x1 convolution with stride 1
	// and no padding, so flag for skipping the buffer and transformation.
	is_1x1_ = true;
	for (int i = 0; i < num_spatial_axes_; ++i) {
		is_1x1_ &=
				kernel_shape_data[i] == 1 && stride_data[i] == 1 && pad_data[i] == 0;
		if (!is_1x1_) { break; }
	}
	// Configure output channels and groups.
	channels_ = bottom[0]->shape(channel_axis_);
	num_output_ = this->layer_param_.weighted_l2_convolution_param().num_output();
	CHECK_GT(num_output_, 0);
	group_ = this->layer_param_.weighted_l2_convolution_param().group();
	CHECK_EQ(channels_ % group_, 0);
	CHECK_EQ(num_output_ % group_, 0)
	<< "Number of output should be multiples of group.";
	if (reverse_dimensions()) {
		conv_out_channels_ = channels_;
		conv_in_channels_ = num_output_;
	} else {
		conv_out_channels_ = num_output_;
		conv_in_channels_ = channels_;
	}
	// Handle the parameters: scales and biases.
	// - blobs_[0] holds the means
	// - blobs_[1] holds the scales
	// - blobs_[2] holds the biases (optional)
	vector<int> mean_scale_shape(2);
	mean_scale_shape[0] = conv_out_channels_;
	mean_scale_shape[1] = conv_in_channels_ / group_;
	for (int i = 0; i < num_spatial_axes_; ++i) {
		mean_scale_shape.push_back(kernel_shape_data[i]);
	}
	bias_term_ = this->layer_param_.weighted_l2_convolution_param().bias_term();
	vector<int> bias_shape(bias_term_, num_output_);
	if (this->blobs_.size() > 0) {
		CHECK_EQ(1 + bias_term_, this->blobs_.size())
        																												<< "Incorrect number of scale blobs.";
		if (mean_scale_shape != this->blobs_[0]->shape()) {
			Blob<Dtype> mean_scale_shaped_blob(mean_scale_shape);
			LOG(FATAL) << "Incorrect mean shape: expected shape "
					<< mean_scale_shaped_blob.shape_string() << "; instead, shape was "
					<< this->blobs_[0]->shape_string();
		}
		if (mean_scale_shape != this->blobs_[1]->shape()) {
			Blob<Dtype> mean_scale_shaped_blob(mean_scale_shape);
			LOG(FATAL) << "Incorrect scale shape: expected shape "
					<< mean_scale_shaped_blob.shape_string() << "; instead, shape was "
					<< this->blobs_[1]->shape_string();
		}
		if (bias_term_ && bias_shape != this->blobs_[2]->shape()) {
			Blob<Dtype> bias_shaped_blob(bias_shape);
			LOG(FATAL) << "Incorrect bias shape: expected shape "
					<< bias_shaped_blob.shape_string() << "; instead, shape was "
					<< this->blobs_[2]->shape_string();
		}
		LOG(INFO) << "Skipping parameter initialization";
	} else {
		if (bias_term_) {
			this->blobs_.resize(3);
		} else {
			this->blobs_.resize(2);
		}
		// Initialize and fill the means:
		// output channels x input channels per-group x kernel height x kernel width
		this->blobs_[0].reset(new Blob<Dtype>(mean_scale_shape));
		shared_ptr<Filler<Dtype> > mean_filler(GetFiller<Dtype>(
				this->layer_param_.weighted_l2_convolution_param().mean_filler()));
		mean_filler->Fill(this->blobs_[0].get());

		// Initialize and fill the scales:
		// output channels x input channels per-group x kernel height x kernel width
		this->blobs_[1].reset(new Blob<Dtype>(mean_scale_shape));
		shared_ptr<Filler<Dtype> > scale_filler(GetFiller<Dtype>(
				this->layer_param_.weighted_l2_convolution_param().scale_filler()));
		scale_filler->Fill(this->blobs_[1].get());
		// If necessary, initialize and fill the biases.
		if (bias_term_) {
			this->blobs_[2].reset(new Blob<Dtype>(bias_shape));
			shared_ptr<Filler<Dtype> > bias_filler(GetFiller<Dtype>(
					this->layer_param_.weighted_l2_convolution_param().bias_filler()));
			bias_filler->Fill(this->blobs_[2].get());
		}
	}
	kernel_dim_ = this->blobs_[0]->count(1);
	weight_offset_ = conv_out_channels_ * kernel_dim_ / group_;
	rev_fan_in_ = conv_out_channels_ * kernel_shape_data[0] * kernel_shape_data[1];
	// Propagate gradients to the parameters (as directed by backward pass).
	this->param_propagate_down_.resize(this->blobs_.size(), true);
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top) {
	const int first_spatial_axis = channel_axis_ + 1;
	CHECK_EQ(bottom[0]->num_axes(), first_spatial_axis + num_spatial_axes_)
	<< "bottom num_axes may not change.";
	num_ = bottom[0]->count(0, channel_axis_);
	CHECK_EQ(bottom[0]->shape(channel_axis_), channels_)
	<< "Input size incompatible with convolution kernel.";
	// TODO: generalize to handle inputs of different shapes.
	for (int bottom_id = 1; bottom_id < bottom.size(); ++bottom_id) {
		CHECK(bottom[0]->shape() == bottom[bottom_id]->shape())        																												<< "All inputs must have the same shape.";
	}
	// Shape the tops.
	bottom_shape_ = &bottom[0]->shape();
	compute_output_shape();
	vector<int> top_shape(bottom[0]->shape().begin(),
			bottom[0]->shape().begin() + channel_axis_);
	top_shape.push_back(num_output_);
	for (int i = 0; i < num_spatial_axes_; ++i) {
		top_shape.push_back(output_shape_[i]);
	}
	for (int top_id = 0; top_id < top.size(); ++top_id) {
		top[top_id]->Reshape(top_shape);
	}
	if (reverse_dimensions()) {
		conv_out_spatial_dim_ = bottom[0]->count(first_spatial_axis);
	} else {
		conv_out_spatial_dim_ = top[0]->count(first_spatial_axis);
	}
	col_offset_ = kernel_dim_ * conv_out_spatial_dim_;
	output_offset_ = conv_out_channels_ * conv_out_spatial_dim_ / group_;
	// Setup input dimensions (conv_input_shape_).
	vector<int> bottom_dim_blob_shape(1, num_spatial_axes_ + 1);
	conv_input_shape_.Reshape(bottom_dim_blob_shape);
	int* conv_input_shape_data = conv_input_shape_.mutable_cpu_data();
	for (int i = 0; i < num_spatial_axes_ + 1; ++i) {
		if (reverse_dimensions()) {
			conv_input_shape_data[i] = top[0]->shape(channel_axis_ + i);
		} else {
			conv_input_shape_data[i] = bottom[0]->shape(channel_axis_ + i);
		}
	}
	// The im2col result buffer will only hold one image at a time to avoid
	// overly large memory usage. In the special case of 1x1 convolution
	// it goes lazily unused to save memory.
	col_buffer_shape_.clear();
	col_buffer_shape_.push_back(kernel_dim_ * group_);
	for (int i = 0; i < num_spatial_axes_; ++i) {
		if (reverse_dimensions()) {
			col_buffer_shape_.push_back(input_shape(i + 1));
		} else {
			col_buffer_shape_.push_back(output_shape_[i]);
		}
	}
	col_buffer_.Reshape(col_buffer_shape_);
	col_buffer_diff_.Reshape(col_buffer_shape_);
	col_buffer_diff_total_.Reshape(col_buffer_shape_);
	rep_mean_buffer_.Reshape (col_buffer_shape_);
	col_mean_diff_sqr_buffer_.Reshape (col_buffer_shape_);
	col_mean_diff_sqr_deriv_buffer_.Reshape (col_buffer_shape_);
	bottom_dim_ = bottom[0]->count(channel_axis_);
	top_dim_ = top[0]->count(channel_axis_);
	num_kernels_im2col_ = conv_in_channels_ * conv_out_spatial_dim_;
	num_kernels_col2im_ = reverse_dimensions() ? top_dim_ : bottom_dim_;
	// Set up the all ones "bias multiplier" for adding biases by BLAS
	out_spatial_dim_ = top[0]->count(first_spatial_axis);
	if (bias_term_) {
		vector<int> bias_multiplier_shape(1, out_spatial_dim_);
		bias_multiplier_.Reshape(bias_multiplier_shape);
		caffe_set(bias_multiplier_.count(), Dtype(1),
				bias_multiplier_.mutable_cpu_data());
	}
	set_blobs(conv_out_channels_ ,conv_out_spatial_dim_,kernel_dim_); // setup blobs according to layer dimensions
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::set_weights_cpu(const Dtype* means,const Dtype* scales){

	//	We computing fixed parameters along iterations in order to obtain efficient computations;
	int M=conv_out_channels_/group_;
	int K=kernel_dim_;
	for (int g=0;g<group_;g++){

		if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_LINEAR) // if scale is linear: scale=scale
		{
			caffe_copy<Dtype>(processed_scales.count()/group_,scales+ weight_offset_ * g,processed_scales.mutable_cpu_data()+ weight_offset_ * g );
		}
		else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_SQUARE)//scale is quadratic: scale=scale^2
		{
			caffe_sqr<Dtype>(processed_scales.count()/group_, scales+ weight_offset_ * g, processed_scales.mutable_cpu_data()+ weight_offset_ * g);
		}
		else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_ABS)//scale is abstract: scale=abs(scale)
		{
			caffe_abs<Dtype>(processed_scales.count()/group_, scales+ weight_offset_ * g, processed_scales.mutable_cpu_data()+ weight_offset_ * g);
			caffe_sign<Dtype>(M*K,scales+weight_offset_ * g,backward_sign_scale.mutable_cpu_data()+weight_offset_ * g);  // we need the sign of the scales in backward phase
		}

		// computing M S
		caffe_mul<Dtype>(M*K,processed_scales.cpu_data()+ weight_offset_ * g,means+ weight_offset_ * g,s_m_dot.mutable_cpu_data()+ weight_offset_ * g); // Hadamard product of scale and mean


		// computing M S M
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, M, K,
				(Dtype)1.0, s_m_dot.cpu_data() + weight_offset_ * g , means+ weight_offset_ * g ,
				(Dtype)0.,forward_temp_1.mutable_cpu_data() );					  // Hadamard product of scale mean^2
		caffe_get_diag_cpu<Dtype>(M,forward_temp_1.mutable_cpu_data(),foward_MSM.mutable_cpu_data()+g*M ); // get only diag elements

		// computing M^2
		caffe_mul<Dtype>(M*K,means+weight_offset_ * g,means+weight_offset_ * g,backward_mean_power_2.mutable_cpu_data()+g*weight_offset_);  // computing mean^2
	}
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::forward_cpu_gemm(const Dtype* input, const Dtype* means,
		const Dtype* scales, Dtype* output, bool skip_im2col) {
	const Dtype* col_buff = input;

//	In foward pass we need to compute the following function:
//	WL2Dist(x,parameters) : conv(scale, X^2) + conv( Haddammard_prod(Scale ,Mean^2), 1) - 2 conv(Haddammard_prod(Scale,Mean),X)
//
//	In the following : M: is stands for Mean and col_buff stands for input and S stands for processed scale.
//
	if (!is_1x1_) {
		if (!skip_im2col) {
			conv_im2col_cpu(input, col_buffer_.mutable_cpu_data());
		}
		col_buff = col_buffer_.cpu_data();
	}

	int M=conv_out_channels_/group_;
	int N=conv_out_spatial_dim_;
	int K=kernel_dim_;


	Dtype	pm1 = 0.5;
	if (negate_distance_)
	{
		pm1 = -0.5;
	}

	Dtype	MSM_sign = 1;
	if (zero_mean_)
	{
		MSM_sign = -1;
	}

	for (int g=0;g<group_;g++){


		// computing M S X
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, N, K,
				(Dtype)(-2.*pm1),s_m_dot.mutable_cpu_data()+ weight_offset_ * g , col_buff + col_offset_ * g,
				(Dtype)0., output  + output_offset_ * g);				  //generated M square matrix

		// computing M S M
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, N, 1,
				(Dtype)pm1*MSM_sign*alpha_, foward_MSM.mutable_cpu_data() +g*M, ones_n.mutable_cpu_data() ,
				(Dtype)(1.0), output  + output_offset_ * g);             // in order to generate M x N matrix

		// computing X S X
		caffe_sqr<Dtype>(N*K, col_buff+ col_offset_ * g, input_power_2.mutable_cpu_data()+ col_offset_ * g); //x^2
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans,M, N, K,
				(Dtype)pm1*alpha_, processed_scales.cpu_data()+ weight_offset_ * g , input_power_2.cpu_data()+ col_offset_ * g,
				(Dtype)1.0, output  + output_offset_ * g);
	}
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::forward_cpu_bias(Dtype* output,
		const Dtype* bias) {
	caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, num_output_,
			out_spatial_dim_, 1, (Dtype)1., bias, bias_multiplier_.cpu_data(),
			(Dtype)1., output);
}


template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::backward_cpu(const Dtype* input,
		const Dtype* output_diff, const Dtype* means, const Dtype* scales, Dtype* input_diff,
		Dtype* means_diff, Dtype* scales_diff) {


//	We compute backward gradient with respect to Input, Scale and Mean parameters.
//

	const Dtype*	col_buff;
	Dtype*	col_buff_diff_total;
	col_buff = input;
	if (!is_1x1_) {
		conv_im2col_cpu(input, col_buffer_.mutable_cpu_data());
		col_buff = col_buffer_.cpu_data();
	}
	Dtype	pm1 = 0.5;
	Dtype	MSM_sign = 1;
	if (negate_distance_)
	{
		pm1 = -0.5;
	}
	if (zero_mean_)
	{
		MSM_sign = -1;
	}


	col_buff_diff_total = col_buffer_diff_total_.mutable_cpu_data();
	caffe_set(col_buffer_diff_total_.count(), Dtype(0), col_buff_diff_total);

	int M=conv_out_channels_/group_;
	int N=conv_out_spatial_dim_;
	int K=kernel_dim_;

	for(int g=0;g<group_;g++){
		Dtype* top_diff_sum_out=backward_temp_1.mutable_cpu_data();

		caffe_cpu_gemv<Dtype>(CblasNoTrans, M,N,
				(Dtype)1.0, output_diff + output_offset_ * g, ones_n.cpu_data() ,
				(Dtype)0., top_diff_sum_out) ;
		caffe_diag_cpu<Dtype>(M,top_diff_sum_out,backward_temp_2.mutable_cpu_data()); // square matrix of summed out backward gradients
		// dx=2aSX -2M.S
		// SX
		Dtype* s_t=backward_temp_3.mutable_cpu_data();
		caffe_cpu_gemm<Dtype>(CblasTrans, CblasNoTrans, K, N, M,
				(Dtype)2.0 * alpha_, processed_scales.cpu_data()+ weight_offset_ * g, 
				output_diff + output_offset_ * g , (Dtype)0., s_t) ;
		caffe_mul(N*K,s_t,col_buff+ col_offset_ * g,col_buff_diff_total+ col_offset_ * g); //dx += 2a(Sxtop_diff)X
		// M.S
		caffe_cpu_gemm<Dtype>(CblasTrans, CblasNoTrans, K, N, M,
				(Dtype)(-2*pm1), s_m_dot.cpu_data()+ weight_offset_ * g , output_diff+ output_offset_ * g ,
				(Dtype)pm1, col_buff_diff_total+ col_offset_ * g) ;  					  //dx -= 2(Hadammard_prod(S,M)xtop_diff)
		// end of dx

		Dtype* x_t_dot=backward_temp_4.mutable_cpu_data();
		caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, K, N,
				(Dtype)1.0, output_diff+ output_offset_ * g, col_buff + col_offset_ * g,
				(Dtype)0., x_t_dot) ;
		//ds
		if (this->param_propagate_down_[1])
		{
			Dtype* temp_scales_diff=backward_temp_5.mutable_cpu_data();
			// ds=  ax^2 + aM^2 -2 * M X
			// M*X
			caffe_mul(M*K,x_t_dot,means+weight_offset_ * g,temp_scales_diff);  // ds+= -2Haddamard_prod(X x top_diff,mean)
			// +ax^2 -2MX
			Dtype* x_x=backward_temp_3.mutable_cpu_data();
			caffe_mul<Dtype>(N*K,col_buff+ col_offset_ * g,col_buff+ col_offset_ * g,x_x); // x^2
			caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, K, N,
					(Dtype)alpha_, output_diff+ output_offset_ * g, x_x ,
					(Dtype)(-2.0), temp_scales_diff) ;				// ds+= alpha X^2 x top_diff
			// +M^2
			const Dtype* m_m=backward_mean_power_2.cpu_data()+g*weight_offset_;
			caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, K, M,
					(Dtype)pm1*MSM_sign*alpha_, backward_temp_2.cpu_data(), m_m ,
					(Dtype)pm1, temp_scales_diff) ;				 // m^2 x sum_out(top_diff)
			if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_LINEAR)
			{
				caffe_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
			else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_SQUARE)
			{
				// ds = 2 S x ds
				caffe_scal<Dtype>(M*K,(Dtype)2.,temp_scales_diff);
				caffe_mul<Dtype>(M*K,scales+weight_offset_ * g,temp_scales_diff,temp_scales_diff);
				caffe_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
			else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_ABS)
			{
				// ds =  sgn(S) x ds
				const Dtype* signs=backward_sign_scale.cpu_data()+weight_offset_ * g;
				caffe_mul<Dtype>(M*K,signs,temp_scales_diff,temp_scales_diff);
				caffe_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
		}

		// end of ds

		// dm = 2* S.M - 2 * S X
		if (this->param_propagate_down_[0])
		{
			// S X
			Dtype* temp_means_diff=backward_temp_6.mutable_cpu_data();
			caffe_mul<Dtype>(M*K,x_t_dot,processed_scales.cpu_data()+weight_offset_ * g,temp_means_diff); // dm += -2Haddamard_prod(scale, X x top_diff)
			// 2 S M
			caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, K, M,
					(Dtype)(2.0*pm1*MSM_sign*alpha_), backward_temp_2.cpu_data(), s_m_dot.cpu_data()+weight_offset_ * g ,
					(Dtype)(-2.*pm1), temp_means_diff) ;     //Haddamard_prod(S,M) x sum_out(top_diff)
			caffe_add<Dtype>(M*K,temp_means_diff,means_diff+weight_offset_ * g,means_diff+weight_offset_ * g);
		}
		// end of dm

	}
	if (natural_gradient_)
		caffe_scal<Dtype>(col_buffer_diff_total_.count(), (Dtype)(1.0/sqrt(rev_fan_in_) ), col_buff_diff_total);  //scale the propagated gradient
	/// new
	if (!is_1x1_) {
		conv_col2im_cpu(col_buff_diff_total, input_diff);
	}
	else
	{
		caffe_copy(col_buffer_diff_total_.count(), col_buff_diff_total, input_diff);
	}
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::backward_cpu_bias(Dtype* bias,
		const Dtype* input) {
	caffe_cpu_gemv<Dtype>(CblasNoTrans, num_output_, out_spatial_dim_, 1.,
			input, bias_multiplier_.cpu_data(), 1., bias);
}

#ifndef CPU_ONLY


template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::set_weights_gpu(const Dtype* means,const Dtype* scales){

	//	We computing fixed parameters along iterations in order to obtain efficient computations;
	int M=conv_out_channels_/group_;
	int K=kernel_dim_;
	for (int g=0;g<group_;g++){

		if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_LINEAR) // if scale is linear: scale=scale
		{
			caffe_copy<Dtype>(processed_scales.count()/group_,scales+ weight_offset_ * g,processed_scales.mutable_gpu_data()+ weight_offset_ * g );
		}
		else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_SQUARE)//scale is quadratic: scale=scale^2
		{
			caffe_gpu_sqr<Dtype>(processed_scales.count()/group_, scales+ weight_offset_ * g, processed_scales.mutable_gpu_data()+ weight_offset_ * g);
		}
		else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_ABS)//scale is abstract: scale=abs(scale)
		{
			caffe_gpu_abs<Dtype>(processed_scales.count()/group_, scales+ weight_offset_ * g, processed_scales.mutable_gpu_data()+ weight_offset_ * g);
			caffe_gpu_sign<Dtype>(M*K,scales+weight_offset_ * g,backward_sign_scale.mutable_gpu_data()+weight_offset_ * g);// we need the sign of the scales in backward phase
		}

		// computing M S
		caffe_gpu_mul<Dtype>(M*K,processed_scales.gpu_data()+ weight_offset_ * g,means+ weight_offset_ * g,s_m_dot.mutable_gpu_data()+ weight_offset_ * g); // Hadamard product of scale and mean


		// computing M S M
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, M, K,
				(Dtype)1.0, s_m_dot.gpu_data() + weight_offset_ * g , means+ weight_offset_ * g ,
				(Dtype)0.,forward_temp_1.mutable_gpu_data() );		// Hadamard product of scale mean^2
		caffe_get_diag_gpu<Dtype>(M,forward_temp_1.mutable_gpu_data(),foward_MSM.mutable_gpu_data()+g*M ); // get only diag elements

		// computing M^2
		caffe_gpu_mul<Dtype>(M*K,means+weight_offset_ * g,means+weight_offset_ * g,backward_mean_power_2.mutable_gpu_data()+g*weight_offset_);   // computing mean^2
	}
}
template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::forward_gpu_gemm(const Dtype* input, const Dtype* means,
		const Dtype* scales, Dtype* output, bool skip_im2col) {
	const Dtype* col_buff = input;

	//	In foward pass we need to compute the following function:
	//	WL2Dist(x,parameters) : conv(scale, X^2) + conv( Haddammard_prod(Scale ,Mean^2), 1) - 2 conv(Haddammard_prod(Scale,Mean),X)
	//
	//	In the following : M: is stands for Mean and col_buff stands for input and S stands for processed scale.
	//

	if (!is_1x1_) {
		if (!skip_im2col) {
			conv_im2col_gpu(input, col_buffer_.mutable_gpu_data());
		}
		col_buff = col_buffer_.gpu_data();
	}

	int M=conv_out_channels_/group_;
	int N=conv_out_spatial_dim_;
	int K=kernel_dim_;


	Dtype	pm1 = 0.5;
	if (negate_distance_)
	{
		pm1 = -0.5;
	}
	Dtype	MSM_sign = 1;
	if (zero_mean_)
	{
		MSM_sign = -1;
	}

	for (int g=0;g<group_;g++){


		// computing M S X
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, N, K,
				(Dtype)(-2.0*pm1),s_m_dot.mutable_gpu_data()+ weight_offset_ * g , col_buff + col_offset_ * g,
				(Dtype)0., output  + output_offset_ * g);//generated M square matrix

		// computing M S M
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, N, 1,
				(Dtype)pm1*MSM_sign*alpha_, foward_MSM.mutable_gpu_data() +g*M, ones_n.mutable_gpu_data() ,
				(Dtype)(1.0), output  + output_offset_ * g);//to generate M x N matrix

		// computing X S X
		caffe_gpu_sqr<Dtype>(N*K, col_buff+ col_offset_ * g, input_power_2.mutable_gpu_data()+ col_offset_ * g); //x^2
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans,M, N, K,
				(Dtype)pm1*alpha_, processed_scales.gpu_data()+ weight_offset_ * g , input_power_2.gpu_data()+ col_offset_ * g,
				(Dtype)1.0, output  + output_offset_ * g);
	}
}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::forward_gpu_bias(Dtype* output,
		const Dtype* bias) {
	caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, num_output_,
			out_spatial_dim_, 1, (Dtype)1., bias, bias_multiplier_.gpu_data(),
			(Dtype)1., output);

}

template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::backward_gpu(const Dtype* input,
		const Dtype* output_diff, const Dtype* means, const Dtype* scales, Dtype* input_diff,
		Dtype* means_diff, Dtype* scales_diff) {
	//	We compute backward gradient with respect to Input, Scale and Mean parameters.

	const Dtype*	col_buff;
	Dtype*	col_buff_diff_total;
	col_buff = input;
	if (!is_1x1_) {
		conv_im2col_gpu(input, col_buffer_.mutable_gpu_data());
		col_buff = col_buffer_.gpu_data();
	}
	Dtype	pm1 = 0.5;
	if (negate_distance_)
	{
		pm1 = -0.5;
	}
	Dtype	MSM_sign = 1;
	if (zero_mean_)
	{
		MSM_sign = -1;
	}

	col_buff_diff_total = col_buffer_diff_total_.mutable_gpu_data();
	caffe_gpu_set(col_buffer_diff_total_.count(), Dtype(0), col_buff_diff_total);


	int M=conv_out_channels_/group_;
	int N=conv_out_spatial_dim_;
	int K=kernel_dim_;

	for(int g=0;g<group_;g++){
		Dtype* top_diff_sum_out=backward_temp_1.mutable_gpu_data();

		caffe_gpu_gemv<Dtype>(CblasNoTrans, M,N,
				(Dtype)1.0, output_diff + output_offset_ * g, ones_n.gpu_data() ,
				(Dtype)0., top_diff_sum_out) ;
		caffe_diag_gpu<Dtype>(M,top_diff_sum_out,backward_temp_2.mutable_gpu_data()); // square matrix of summed out backward gradients
		// dx=2aSX -2M.S
		// SX
		Dtype* s_t=backward_temp_3.mutable_gpu_data();
		caffe_gpu_gemm<Dtype>(CblasTrans, CblasNoTrans, K, N, M,
				(Dtype)2.0*alpha_, processed_scales.gpu_data()+ weight_offset_ * g, output_diff + output_offset_ * g ,
				(Dtype)0., s_t) ;
		caffe_gpu_mul(N*K,s_t,col_buff+ col_offset_ * g,col_buff_diff_total+ col_offset_ * g); //dx += 2a(Sxtop_diff)X
		// M.S
		caffe_gpu_gemm<Dtype>(CblasTrans, CblasNoTrans, K, N, M,
				(Dtype)(-2*pm1), s_m_dot.gpu_data()+ weight_offset_ * g , output_diff+ output_offset_ * g ,
				(Dtype)pm1, col_buff_diff_total+ col_offset_ * g) ;					//dx -= 2(Hadammard_prod(S,M)xtop_diff)
		// end of dx

		Dtype* x_t_dot=backward_temp_4.mutable_gpu_data();
		caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, K, N,
				(Dtype)1.0, output_diff+ output_offset_ * g, col_buff + col_offset_ * g,
				(Dtype)0., x_t_dot) ;
		//ds
		if (this->param_propagate_down_[1])
		{
			Dtype* temp_scales_diff=backward_temp_5.mutable_gpu_data();
			// ds=  x^2 + M^2 -2 * M X
			// M*X
			caffe_gpu_mul(M*K,x_t_dot,means+weight_offset_ * g,temp_scales_diff);  // ds+= -2Haddamard_prod(X x top_diff,mean)
			// +ax^2 -2MX
			Dtype* x_x=backward_temp_3.mutable_gpu_data();
			caffe_gpu_mul<Dtype>(N*K,col_buff+ col_offset_ * g,col_buff+ col_offset_ * g,x_x);
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M, K, N,
					(Dtype)alpha_, output_diff+ output_offset_ * g, x_x ,
					(Dtype)(-2.0), temp_scales_diff) ; 		// ds+= alpha X^2 x top_diff
			// +M^2
			const Dtype* m_m=backward_mean_power_2.gpu_data()+g*weight_offset_;
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, K, M,
					(Dtype)pm1*MSM_sign*alpha_, backward_temp_2.gpu_data(), m_m ,
					(Dtype)pm1, temp_scales_diff) ;			// m^2 x sum_out(top_diff)
			if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_LINEAR)
			{
				caffe_gpu_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
			else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_SQUARE)
			{
				// ds = 2 S ds
				caffe_gpu_scal<Dtype>(M*K,(Dtype)2.,temp_scales_diff);
				caffe_gpu_mul<Dtype>(M*K,scales+weight_offset_ * g,temp_scales_diff,temp_scales_diff);
				caffe_gpu_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
			else if (scale_parameterization_ == WeightedL2ConvolutionParameter_ScaleParameterization_SP_ABS)
			{
				// ds =  sgn(S) ds
				const Dtype* signs=backward_sign_scale.gpu_data()+weight_offset_ * g;
				caffe_gpu_mul<Dtype>(M*K,signs,temp_scales_diff,temp_scales_diff);
				caffe_gpu_add<Dtype>(M*K,temp_scales_diff,scales_diff+weight_offset_ * g,scales_diff+weight_offset_ * g);
			}
		}

		// end of ds

		// dm = 2* S.M - 2 * S X
		if (this->param_propagate_down_[0])
		{
			// S X
			Dtype* temp_means_diff=backward_temp_6.mutable_gpu_data();
			caffe_gpu_mul<Dtype>(M*K,x_t_dot,processed_scales.gpu_data()+weight_offset_ * g,temp_means_diff); 	// dm += -2Haddamard_prod(scale, X x top_diff)
			// 2 S M
			caffe_gpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M, K, M,
					(Dtype)(2.0*pm1*MSM_sign*alpha_), backward_temp_2.gpu_data(), s_m_dot.gpu_data()+weight_offset_ * g ,
					(Dtype)(-2.*pm1), temp_means_diff) ;		//Haddamard_prod(S,M) x sum_out(top_diff)
			caffe_gpu_add<Dtype>(M*K,temp_means_diff,means_diff+weight_offset_ * g,means_diff+weight_offset_ * g);
		}
		// end of dm

	}
	if (natural_gradient_)
		caffe_gpu_scal<Dtype>(col_buffer_diff_total_.count(), (Dtype)(1.0/sqrt(rev_fan_in_) ), col_buff_diff_total);  //scale the propagated gradient
	/// new
	if (!is_1x1_) {
		conv_col2im_gpu(col_buff_diff_total, input_diff);
	}
	else
	{
		caffe_copy(col_buffer_diff_total_.count(), col_buff_diff_total, input_diff);
	}
}



template <typename Dtype>
void BaseWeightedL2ConvolutionLayer<Dtype>::backward_gpu_bias(Dtype* bias,
		const Dtype* input) {
	caffe_gpu_gemv<Dtype>(CblasNoTrans, num_output_, out_spatial_dim_, 1.,
			input, bias_multiplier_.gpu_data(), 1., bias);
}

#endif  // !CPU_ONLY

INSTANTIATE_CLASS(BaseWeightedL2ConvolutionLayer);

}  // namespace caffe
