// Authors:   Keivan Nalaie 
//            Kamaledin Ghiasi-Shirazi

#include <vector>
#include "caffe/layers/weighted_l2_convolution_layer.hpp"

namespace caffe {

template <typename Dtype>
void WeightedL2ConvolutionLayer<Dtype>::compute_output_shape() {
  const int* kernel_shape_data = this->kernel_shape_.gpu_data();
  const int* stride_data = this->stride_.gpu_data();
  const int* pad_data = this->pad_.gpu_data();
  const int* dilation_data = this->dilation_.gpu_data();
  this->output_shape_.clear();
  for (int i = 0; i < this->num_spatial_axes_; ++i) {
    // i + 1 to skip channel axis
    const int input_dim = this->input_shape(i + 1);
    const int kernel_extent = dilation_data[i] * (kernel_shape_data[i] - 1) + 1;
    const int output_dim = (input_dim + 2 * pad_data[i] - kernel_extent)
        / stride_data[i] + 1;
    this->output_shape_.push_back(output_dim);
  }
}

template <typename Dtype>
void WeightedL2ConvolutionLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  const Dtype* means = this->blobs_[0]->gpu_data();
  const Dtype* scales = this->blobs_[1]->gpu_data();
  this->set_weights_gpu(means,scales);    // compute parameters which are fixed in each iteration
  for (int i = 0; i < bottom.size(); ++i) {
	  const Dtype* bottom_data = bottom[i]->gpu_data();
	  Dtype* top_data_diff = top[i]->mutable_gpu_data();
	  for (int n = 0; n < this->num_; ++n) {
		  this->forward_gpu_gemm(bottom_data + n * this->bottom_dim_, means, scales,
				  top_data_diff + n * this->top_dim_);

		  if (this->bias_term_) {
			  const Dtype* bias = this->blobs_[2]->gpu_data();
			  this->forward_gpu_bias(top_data_diff + n * this->top_dim_, bias);
      }
    }
  }
}

template <typename Dtype>
void WeightedL2ConvolutionLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
      const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {
  const Dtype* mean = this->blobs_[0]->gpu_data();
  Dtype* mean_diff = this->blobs_[0]->mutable_gpu_diff();
  const Dtype* scale = this->blobs_[1]->gpu_data();

  Dtype* scale_diff = this->blobs_[1]->mutable_gpu_diff();  
  for (int i = 0; i < top.size(); ++i) {
    const Dtype* top_diff = top[i]->gpu_diff();
    const Dtype* bottom_data = bottom[i]->gpu_data();
    Dtype* bottom_diff = bottom[i]->mutable_gpu_diff();
    // Bias gradient, if necessary.
    if (this->bias_term_ && this->param_propagate_down_[2]) {
      Dtype* bias_diff = this->blobs_[2]->mutable_gpu_diff();
      for (int n = 0; n < this->num_; ++n) {
        this->backward_gpu_bias(bias_diff, top_diff + n * this->top_dim_);
      }
    }
    if (this->param_propagate_down_[0] || this->param_propagate_down_[1] || propagate_down[i]) {
      for (int n = 0; n < this->num_; ++n) {
    	  // gradient w.r.t. mean and scale. Note that we will accumulate diffs.
    	    this->backward_gpu(bottom_data + n * this->bottom_dim_,
    	    		top_diff + n * this->top_dim_, mean, scale, bottom_diff + n * this->bottom_dim_, mean_diff, scale_diff);

        }
    }
  }
}


INSTANTIATE_LAYER_GPU_FUNCS(WeightedL2ConvolutionLayer);

}  // namespace caffe
