#include <cstring>
#include <vector>

#include "gtest/gtest.h"

#include "caffe/blob.hpp"
#include "caffe/common.hpp"
#include "caffe/filler.hpp"
#include "caffe/layers/euclidean_convolution_layer.hpp"

#include "caffe/test/test_caffe_main.hpp"
#include "caffe/test/test_gradient_check_util.hpp"

namespace caffe {

// Reference Euclidean Similarity for checking results:
// accumulate through explicit loops over input, output, and filters.
template <typename Dtype>
void caffe_euc_sim(const Blob<Dtype>* in, EuclideanSimilarityParameter* euc_sim_param,
                   const vector<shared_ptr<Blob<Dtype> > >& blobs,
                   Blob<Dtype>* out) {
    // Kernel size, stride, and pad
    int kernel_h, kernel_w;
    Dtype normConst;
    Dtype additiveTerm;
    EuclideanSimilarityParameter_MeanParameterization mean_parameterization_;
    EuclideanSimilarityParameter_ScaleParameterization scale_parameterization_;
    EuclideanSimilarityParameter_NormalizationMethod   normalization_;
    
    if (euc_sim_param->has_kernel_size()) {
        kernel_h = kernel_w = euc_sim_param->kernel_size();
    } else {
        kernel_h = euc_sim_param->kernel_h();
        kernel_w = euc_sim_param->kernel_w();
    }
    int pad_h, pad_w;
    if (!euc_sim_param->has_pad_h()) {
        pad_h = pad_w = euc_sim_param->pad();
    } else {
        pad_h = euc_sim_param->pad_h();
        pad_w = euc_sim_param->pad_w();
    }
    int stride_h, stride_w;
    if (!euc_sim_param->has_stride_h()) {
        stride_h = stride_w = euc_sim_param->stride();
    } else {
        stride_h = euc_sim_param->stride_h();
        stride_w = euc_sim_param->stride_w();
    }
    mean_parameterization_  = euc_sim_param->mean_parameterization();
    scale_parameterization_ = euc_sim_param->scale_parameterization();
   
    // Groups
    int groups = euc_sim_param->group();
    int o_g = out->channels() / groups;
    int k_g = in->channels() / groups;
    int o_head, k_head;
    // Euclidean Similarity
    const Dtype* in_data = in->cpu_data();
    const Dtype* mean_data = blobs[0]->cpu_data();
    const Dtype* scale_data = blobs[1]->cpu_data();
    Dtype* out_data = out->mutable_cpu_data();
    Dtype s, positive_scale=0, positive_mean=0;
    //Dtype normConst = Dtype(sqrt(kernel_h * kernel_w * k_g));
    //Dtype additiveTerm = Dtype(-sqrt(kernel_h * kernel_w * k_g));
    
    normalization_  = euc_sim_param->normalization();
    if (normalization_ == EuclideanSimilarityParameter_NormalizationMethod_NONE)
    {
      normConst = 1;
      additiveTerm = 0;    
    }
    else if (normalization_ == EuclideanSimilarityParameter_NormalizationMethod_MEAN_ZERO_VAR_TWO)
    {
      normConst = Dtype(sqrt(kernel_h * kernel_w * k_g));
      additiveTerm = Dtype(-sqrt(kernel_h * kernel_w * k_g));
    }
    else
    {
	LOG(ERROR) << "Invalid Normalization Method";
    }    
    
    
    
    for (int n = 0; n < out->num(); n++) {
        for (int g = 0; g < groups; g++) {
            o_head = o_g * g;
            k_head = k_g * g;
            for (int o = 0; o < o_g; o++) {
                for (int y = 0; y < out->height(); y++) {
                    for (int x = 0; x < out->width(); x++) {
			for (int k = 0; k < k_g; k++) {
                            for (int p = 0; p < kernel_h; p++) {
                                for (int q = 0; q < kernel_w; q++) {
                                    int in_y = y * stride_h - pad_h + p;
                                    int in_x = x * stride_w - pad_w + q;
                                    if (in_y >= 0 && in_y < in->height()
                                            && in_x >= 0 && in_x < in->width()) {
				        if (scale_parameterization_ == EuclideanSimilarityParameter_ScaleParameterization_SP_LINEAR)
					{
                                            positive_scale = scale_data[blobs[0]->offset(o + o_head, k, p, q)];					  
					}
				        else if (scale_parameterization_ == EuclideanSimilarityParameter_ScaleParameterization_SP_ABS)
					{
                                            positive_scale = fabs(scale_data[blobs[0]->offset(o + o_head, k, p, q)]);
					}
                                        else if (scale_parameterization_ == EuclideanSimilarityParameter_ScaleParameterization_SP_QUADRATIC)
                                        {
                                            positive_scale = scale_data[blobs[0]->offset(o + o_head, k, p, q)];
                                            positive_scale *= positive_scale;
                                        }
                                        else if (scale_parameterization_ == EuclideanSimilarityParameter_ScaleParameterization_SP_EXPONENTIAL)
                                        {
                                            positive_scale = exp(scale_data[blobs[0]->offset(o + o_head, k, p, q)]);
                                        }

					positive_mean = mean_data[blobs[0]->offset(o + o_head, k, p, q)];
                                        if (mean_parameterization_ == EuclideanSimilarityParameter_MeanParameterization_MP_QUADRATIC)
					{
					  positive_mean *= positive_mean;
					}
					else if (mean_parameterization_ == EuclideanSimilarityParameter_MeanParameterization_MP_SIGMOID)
					{
					  positive_mean = 1.0 / (1.0 + exp(-positive_mean));
					}
										
                                        s = (in_data[in->offset(n, k + k_head, in_y, in_x)]- positive_mean);
                                        out_data[out->offset(n, o + o_head, y, x)] += s * s * positive_scale;
                                    }
                                }
                            }
                        }
                        out_data[out->offset(n, o + o_head, y, x)] = out_data[out->offset(n, o + o_head, y, x)]/normConst + additiveTerm;
                    }
                }
                
            }
        }
    }
}

template void caffe_euc_sim(const Blob<float>* in,
                            EuclideanSimilarityParameter* euc_sim_param,
                            const vector<shared_ptr<Blob<float> > >& weights,
                            Blob<float>* out);
template void caffe_euc_sim(const Blob<double>* in,
                            EuclideanSimilarityParameter* euc_sim_param,
                            const vector<shared_ptr<Blob<double> > >& weights,
                            Blob<double>* out);

template <typename TypeParam>
class EuclideanSimilarityLayerTest : public MultiDeviceTest<TypeParam> {
    typedef typename TypeParam::Dtype Dtype;

protected:
    EuclideanSimilarityLayerTest()
        : blob_bottom_(new Blob<Dtype>(2, 3, 6, 4)),
          blob_bottom_2_(new Blob<Dtype>(2, 3, 6, 4)),
          blob_top_(new Blob<Dtype>()),
          blob_top_2_(new Blob<Dtype>()) {}
    virtual void SetUp() {
        // fill the values
        FillerParameter filler_param;
        filler_param.set_value(1.);
        //GaussianFiller<Dtype> filler(filler_param);
        ConstantFiller<Dtype> filler(filler_param);
        filler.Fill(this->blob_bottom_);
        filler.Fill(this->blob_bottom_2_);
        blob_bottom_vec_.push_back(blob_bottom_);
        blob_top_vec_.push_back(blob_top_);
    }

    virtual ~EuclideanSimilarityLayerTest() {
        delete blob_bottom_;
        delete blob_bottom_2_;
        delete blob_top_;
        delete blob_top_2_;
    }

    virtual Blob<Dtype>* MakeReferenceTop(Blob<Dtype>* top) {
        this->ref_blob_top_.reset(new Blob<Dtype>());
        this->ref_blob_top_->ReshapeLike(*top);
        return this->ref_blob_top_.get();
    }

    Blob<Dtype>* const blob_bottom_;
    Blob<Dtype>* const blob_bottom_2_;
    Blob<Dtype>* const blob_top_;
    Blob<Dtype>* const blob_top_2_;
    shared_ptr<Blob<Dtype> > ref_blob_top_;
    vector<Blob<Dtype>*> blob_bottom_vec_;
    vector<Blob<Dtype>*> blob_top_vec_;
};

TYPED_TEST_CASE(EuclideanSimilarityLayerTest, TestDtypesAndDevices);

TYPED_TEST(EuclideanSimilarityLayerTest, TestSetup) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(4);
    euc_sim_param->set_mean_parameterization(EuclideanSimilarityParameter_MeanParameterization_MP_SIGMOID);
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    shared_ptr<Layer<Dtype> > layer(
        new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    EXPECT_EQ(this->blob_top_->num(), 2);
    EXPECT_EQ(this->blob_top_->channels(), 4);
    EXPECT_EQ(this->blob_top_->height(), 2);
    EXPECT_EQ(this->blob_top_->width(), 1);
    EXPECT_EQ(this->blob_top_2_->num(), 2);
    EXPECT_EQ(this->blob_top_2_->channels(), 4);
    EXPECT_EQ(this->blob_top_2_->height(), 2);
    EXPECT_EQ(this->blob_top_2_->width(), 1);
    // setting group should not change the shape
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    layer.reset(new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    EXPECT_EQ(this->blob_top_->num(), 2);
    EXPECT_EQ(this->blob_top_->channels(), 3);
    EXPECT_EQ(this->blob_top_->height(), 2);
    EXPECT_EQ(this->blob_top_->width(), 1);
    EXPECT_EQ(this->blob_top_2_->num(), 2);
    EXPECT_EQ(this->blob_top_2_->channels(), 3);
    EXPECT_EQ(this->blob_top_2_->height(), 2);
    EXPECT_EQ(this->blob_top_2_->width(), 1);
}

TYPED_TEST(EuclideanSimilarityLayerTest, TestSimpleEuclideanSimilarity) {
    typedef typename TypeParam::Dtype Dtype;
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_normalization(EuclideanSimilarityParameter_NormalizationMethod_NONE);
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(4);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    shared_ptr<Layer<Dtype> > layer(
        new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const Dtype* top_data;
    const Dtype* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
    caffe_euc_sim(this->blob_bottom_2_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_2_));
    top_data = this->blob_top_2_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

TYPED_TEST(EuclideanSimilarityLayerTest, TestSimpleEuclideanSimilarityNormalized) {
    typedef typename TypeParam::Dtype Dtype;
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_normalization(EuclideanSimilarityParameter_NormalizationMethod_MEAN_ZERO_VAR_TWO);
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(4);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    shared_ptr<Layer<Dtype> > layer(
        new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const Dtype* top_data;
    const Dtype* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
    caffe_euc_sim(this->blob_bottom_2_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_2_));
    top_data = this->blob_top_2_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

TYPED_TEST(EuclideanSimilarityLayerTest, Test1x1EuclideanSimilarity) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(1);
    euc_sim_param->set_stride(1);
    euc_sim_param->set_num_output(4);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    shared_ptr<Layer<Dtype> > layer(
        new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const Dtype* top_data;
    const Dtype* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

TYPED_TEST(EuclideanSimilarityLayerTest, TestSimpleEuclideanSimilarityGroup) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    shared_ptr<Layer<Dtype> > layer(
        new EuclideanSimilarityLayer<Dtype>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const Dtype* top_data;
    const Dtype* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

// TYPED_TEST(EuclideanSimilarityLayerTest, TestSobelEuclideanSimilarity) {
//     // Test separable WeightedL2Conv by computing the Sobel operator
//     // as a single filter then comparing the result
//     // as the WeightedL2Conv of two rectangular filters.
//     typedef typename TypeParam::Dtype Dtype;
//     // Fill bottoms with identical Gaussian noise.
//     shared_ptr<GaussianFiller<Dtype> > filler;
//     FillerParameter filler_param;
//     filler_param.set_value(1.);
//     filler.reset(new GaussianFiller<Dtype>(filler_param));
//     filler->Fill(this->blob_bottom_);
//     this->blob_bottom_2_->CopyFrom(*this->blob_bottom_);
//     // Compute Sobel G_x operator as 3 x 3 WeightedL2Conv.
//     LayerParameter layer_param;
//     EuclideanSimilarityParameter* euc_sim_param =
//         layer_param.mutable_euc_sim_param();
//     euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_QUADRATIC);
//     euc_sim_param->set_kernel_size(3);
//     euc_sim_param->set_stride(2);
//     euc_sim_param->set_num_output(1);
//     shared_ptr<Layer<Dtype> > layer(
//         new EuclideanSimilarityLayer<Dtype>(layer_param));
//     layer->blobs().resize(3);
//     layer->blobs()[0].reset(new Blob<Dtype>(1, 3, 3, 3));
//     layer->blobs()[1].reset(new Blob<Dtype>(1, 3, 3, 3));
//     layer->blobs()[2].reset(new Blob<Dtype>(1, 1, 1, 1));
//     Dtype* means = layer->blobs()[0]->mutable_cpu_data();
//     Dtype* scale = layer->blobs()[1]->mutable_cpu_data();
//     Dtype  scale_constant_value = (Dtype)1.0;
// 
//     for (int c = 0; c < 3; ++c) {
//         int i = c * 9;  // 3 x 3 filter
//         means[i +  0] = -1;
//         means[i +  1] =  0;
//         means[i +  2] =  1;
//         means[i +  3] = -2;
//         means[i +  4] =  0;
//         means[i +  5] =  2;
//         means[i +  6] = -1;
//         means[i +  7] =  0;
//         means[i +  8] =  1;
// 
//         scale[i +  0] =  scale_constant_value;
//         scale[i +  1] =  scale_constant_value;
//         scale[i +  2] =  scale_constant_value;
//         scale[i +  3] =  scale_constant_value;
//         scale[i +  4] =  scale_constant_value;
//         scale[i +  5] =  scale_constant_value;
//         scale[i +  6] =  scale_constant_value;
//         scale[i +  7] =  scale_constant_value;
//         scale[i +  8] =  scale_constant_value;
// 
//     }
//     layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
//     layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
//     // Compute Sobel G_x operator as separable 3 x 1 and 1 x 3 WeightedL2Convs.
//     // (1) the [1 2 1] column filter
//     vector<Blob<Dtype>*> sep_blob_bottom_vec;
//     vector<Blob<Dtype>*> sep_blob_top_vec;
//     shared_ptr<Blob<Dtype> > blob_sep(new Blob<Dtype>());
//     sep_blob_bottom_vec.push_back(this->blob_bottom_2_);
//     sep_blob_top_vec.push_back(this->blob_top_2_);
//     euc_sim_param->clear_kernel_size();
//     euc_sim_param->clear_stride();
//     euc_sim_param->set_kernel_h(3);
//     euc_sim_param->set_kernel_w(1);
//     euc_sim_param->set_stride_h(2);
//     euc_sim_param->set_stride_w(1);
//     euc_sim_param->set_num_output(1);
//     layer.reset(new EuclideanSimilarityLayer<Dtype>(layer_param));
//     layer->blobs().resize(3);
//     layer->blobs()[0].reset(new Blob<Dtype>(1, 3, 3, 1));
//     layer->blobs()[1].reset(new Blob<Dtype>(1, 3, 3, 1));
//     layer->blobs()[2].reset(new Blob<Dtype>(1, 1, 1, 1));
//     Dtype* means_1 = layer->blobs()[0]->mutable_cpu_data();
//     Dtype* scale_1 = layer->blobs()[1]->mutable_cpu_data();
//     for (int c = 0; c < 3; ++c) {
//         int i = c * 3;  // 3 x 1 filter
//         means_1[i +  0] = 1;
//         means_1[i +  1] = 2;
//         means_1[i +  2] = 1;
// 
//         scale_1[i +  0] = scale_constant_value;
//         scale_1[i +  1] = scale_constant_value;
//         scale_1[i +  2] = scale_constant_value;
// 
//     }
//     layer->SetUp(sep_blob_bottom_vec, sep_blob_top_vec);
//     layer->Forward(sep_blob_bottom_vec, sep_blob_top_vec);
//     // (2) the [-1 0 1] row filter
//     blob_sep->CopyFrom(*this->blob_top_2_, false, true);
//     sep_blob_bottom_vec.clear();
//     sep_blob_bottom_vec.push_back(blob_sep.get());
//     euc_sim_param->set_kernel_h(1);
//     euc_sim_param->set_kernel_w(3);
//     euc_sim_param->set_stride_h(1);
//     euc_sim_param->set_stride_w(2);
//     euc_sim_param->set_num_output(1);
//     layer.reset(new EuclideanSimilarityLayer<Dtype>(layer_param));
//     layer->blobs().resize(3);
//     layer->blobs()[0].reset(new Blob<Dtype>(1, 3, 1, 3));
//     layer->blobs()[1].reset(new Blob<Dtype>(1, 3, 1, 3));
//     layer->blobs()[2].reset(new Blob<Dtype>(1, 1, 1, 1));
//     Dtype* means_2 = layer->blobs()[0]->mutable_cpu_data();
//     Dtype* scale_2 = layer->blobs()[1]->mutable_cpu_data();
//     for (int c = 0; c < 3; ++c) {
//         int i = c * 3;  // 1 x 3 filter
//         means_2[i +  0] =  1;
//         means_2[i +  1] =  1;
//         means_2[i +  2] =  1;
// 
//         scale_2[i +  0] =  scale_constant_value;
//         scale_2[i +  1] =  scale_constant_value;
//         scale_2[i +  2] =  scale_constant_value;
//     }
//     layer->SetUp(sep_blob_bottom_vec, sep_blob_top_vec);
//     layer->Forward(sep_blob_bottom_vec, sep_blob_top_vec);
//     // Test equivalence of full and separable filters.
//     const Dtype* top_data = this->blob_top_->cpu_data();
//     const Dtype* sep_top_data = this->blob_top_2_->cpu_data();
//     for (int i = 0; i < this->blob_top_->count(); ++i) {
//         EXPECT_NEAR(top_data[i], sep_top_data[i], 1e-4);
//     }
// }

TYPED_TEST(EuclideanSimilarityLayerTest, TestGradient) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    euc_sim_param->set_normalization(EuclideanSimilarityParameter_NormalizationMethod_NONE);
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(2);
    euc_sim_param->mutable_mean_filler()->set_type("constant");
    euc_sim_param->mutable_mean_filler()->set_value(2.0);
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    EuclideanSimilarityLayer<Dtype> layer(layer_param);
    layer.SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    GradientChecker<Dtype> checker(1e-2, 2*1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

TYPED_TEST(EuclideanSimilarityLayerTest, TestGradientNormalized) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    euc_sim_param->set_normalization(EuclideanSimilarityParameter_NormalizationMethod_MEAN_ZERO_VAR_TWO);
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(2);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    EuclideanSimilarityLayer<Dtype> layer(layer_param);
    layer.SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    GradientChecker<Dtype> checker(1e-2, 2*1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

TYPED_TEST(EuclideanSimilarityLayerTest, Test1x1Gradient) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    euc_sim_param->set_kernel_size(1);
    euc_sim_param->set_stride(1);
    euc_sim_param->set_num_output(2);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);

    EuclideanSimilarityLayer<Dtype> layer(layer_param);
    layer.SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    GradientChecker<Dtype> checker(1e-2, 1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

TYPED_TEST(EuclideanSimilarityLayerTest, TestGradientGroup) {
    typedef typename TypeParam::Dtype Dtype;
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    euc_sim_param->set_scale_parameterization(EuclideanSimilarityParameter_ScaleParameterization_SP_ABS);
    EuclideanSimilarityLayer<Dtype> layer(layer_param);
    layer.SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    GradientChecker<Dtype> checker(1e-2, 1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

#ifdef USE_CUDNN

template <typename Dtype>
class CuDNNEuclideanSimilarityLayerTest : public GPUDeviceTest<Dtype> {
protected:
    CuDNNEuclideanSimilarityLayerTest()
        : blob_bottom_(new Blob<Dtype>(2, 3, 6, 4)),
          blob_bottom_2_(new Blob<Dtype>(2, 3, 6, 4)),
          blob_top_(new Blob<Dtype>()),
          blob_top_2_(new Blob<Dtype>()) {}
    virtual void SetUp() {
        // fill the values
        FillerParameter filler_param;
        filler_param.set_value(1.);
        GaussianFiller<Dtype> filler(filler_param);
        filler.Fill(this->blob_bottom_);
        filler.Fill(this->blob_bottom_2_);
        blob_bottom_vec_.push_back(blob_bottom_);
        blob_top_vec_.push_back(blob_top_);
    }

    virtual ~CuDNNEuclideanSimilarityLayerTest() {
        delete blob_bottom_;
        delete blob_bottom_2_;
        delete blob_top_;
        delete blob_top_2_;
    }

    virtual Blob<Dtype>* MakeReferenceTop(Blob<Dtype>* top) {
        this->ref_blob_top_.reset(new Blob<Dtype>());
        this->ref_blob_top_->ReshapeLike(*top);
        return this->ref_blob_top_.get();
    }

    Blob<Dtype>* const blob_bottom_;
    Blob<Dtype>* const blob_bottom_2_;
    Blob<Dtype>* const blob_top_;
    Blob<Dtype>* const blob_top_2_;
    shared_ptr<Blob<Dtype> > ref_blob_top_;
    vector<Blob<Dtype>*> blob_bottom_vec_;
    vector<Blob<Dtype>*> blob_top_vec_;
};

TYPED_TEST_CASE(CuDNNEuclideanSimilarityLayerTest, TestDtypes);

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestSetupCuDNN) {
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(4);
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    shared_ptr<Layer<TypeParam> > layer(
        new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    EXPECT_EQ(this->blob_top_->num(), 2);
    EXPECT_EQ(this->blob_top_->channels(), 4);
    EXPECT_EQ(this->blob_top_->height(), 2);
    EXPECT_EQ(this->blob_top_->width(), 1);
    EXPECT_EQ(this->blob_top_2_->num(), 2);
    EXPECT_EQ(this->blob_top_2_->channels(), 4);
    EXPECT_EQ(this->blob_top_2_->height(), 2);
    EXPECT_EQ(this->blob_top_2_->width(), 1);
    // setting group should not change the shape
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    layer.reset(new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    EXPECT_EQ(this->blob_top_->num(), 2);
    EXPECT_EQ(this->blob_top_->channels(), 3);
    EXPECT_EQ(this->blob_top_->height(), 2);
    EXPECT_EQ(this->blob_top_->width(), 1);
    EXPECT_EQ(this->blob_top_2_->num(), 2);
    EXPECT_EQ(this->blob_top_2_->channels(), 3);
    EXPECT_EQ(this->blob_top_2_->height(), 2);
    EXPECT_EQ(this->blob_top_2_->width(), 1);
}

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestSimpleEuclideanSimilarityCuDNN) {
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(4);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    shared_ptr<Layer<TypeParam> > layer(
        new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const TypeParam* top_data;
    const TypeParam* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
    caffe_euc_sim(this->blob_bottom_2_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_2_));
    top_data = this->blob_top_2_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestSimpleEuclideanSimilarityGroupCuDNN) {
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("constant");
    euc_sim_param->mutable_scale_filler()->set_value(0.1);
    shared_ptr<Layer<TypeParam> > layer(
        new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Check against reference WeightedL2Conv.
    const TypeParam* top_data;
    const TypeParam* ref_top_data;
    caffe_euc_sim(this->blob_bottom_, euc_sim_param, layer->blobs(),
                  this->MakeReferenceTop(this->blob_top_));
    top_data = this->blob_top_->cpu_data();
    ref_top_data = this->ref_blob_top_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], ref_top_data[i], 1e-4);
    }
}

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestSobelEuclideanSimilarityCuDNN) {
    // Test separable WeightedL2Conv by computing the Sobel operator
    // as a single filter then comparing the result
    // as the WeightedL2Conv of two rectangular filters.

    // Fill bottoms with identical Gaussian noise.
    shared_ptr<GaussianFiller<TypeParam> > filler;
    FillerParameter filler_param;
    filler_param.set_value(1.);
    filler.reset(new GaussianFiller<TypeParam>(filler_param));
    filler->Fill(this->blob_bottom_);
    this->blob_bottom_2_->CopyFrom(*this->blob_bottom_);
    // Compute Sobel G_x operator as 3 x 3 WeightedL2Conv.
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(1);
    euc_sim_param->set_bias_term(false);
    shared_ptr<Layer<TypeParam> > layer(
        new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->blobs().resize(1);
    layer->blobs()[0].reset(new Blob<TypeParam>(1, 3, 3, 3));
    TypeParam* weights = layer->blobs()[0]->mutable_cpu_data();
    for (int c = 0; c < 3; ++c) {
        int i = c * 9;  // 3 x 3 filter
        weights[i +  0] = -1;
        weights[i +  1] =  0;
        weights[i +  2] =  1;
        weights[i +  3] = -2;
        weights[i +  4] =  0;
        weights[i +  5] =  2;
        weights[i +  6] = -1;
        weights[i +  7] =  0;
        weights[i +  8] =  1;
    }
    layer->SetUp(this->blob_bottom_vec_, this->blob_top_vec_);
    layer->Forward(this->blob_bottom_vec_, this->blob_top_vec_);
    // Compute Sobel G_x operator as separable 3 x 1 and 1 x 3 WeightedL2Convs.
    // (1) the [1 2 1] column filter
    vector<Blob<TypeParam>*> sep_blob_bottom_vec;
    vector<Blob<TypeParam>*> sep_blob_top_vec;
    shared_ptr<Blob<TypeParam> > blob_sep(new Blob<TypeParam>());
    sep_blob_bottom_vec.push_back(this->blob_bottom_2_);
    sep_blob_top_vec.push_back(this->blob_top_2_);
    euc_sim_param->clear_kernel_size();
    euc_sim_param->clear_stride();
    euc_sim_param->set_kernel_h(3);
    euc_sim_param->set_kernel_w(1);
    euc_sim_param->set_stride_h(2);
    euc_sim_param->set_stride_w(1);
    euc_sim_param->set_num_output(1);
    euc_sim_param->set_bias_term(false);
    layer.reset(new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->blobs().resize(1);
    layer->blobs()[0].reset(new Blob<TypeParam>(1, 3, 3, 1));
    TypeParam* weights_1 = layer->blobs()[0]->mutable_cpu_data();
    for (int c = 0; c < 3; ++c) {
        int i = c * 3;  // 3 x 1 filter
        weights_1[i +  0] = 1;
        weights_1[i +  1] = 2;
        weights_1[i +  2] = 1;
    }
    layer->SetUp(sep_blob_bottom_vec, sep_blob_top_vec);
    layer->Forward(sep_blob_bottom_vec, sep_blob_top_vec);
    // (2) the [-1 0 1] row filter
    blob_sep->CopyFrom(*this->blob_top_2_, false, true);
    sep_blob_bottom_vec.clear();
    sep_blob_bottom_vec.push_back(blob_sep.get());
    euc_sim_param->set_kernel_h(1);
    euc_sim_param->set_kernel_w(3);
    euc_sim_param->set_stride_h(1);
    euc_sim_param->set_stride_w(2);
    euc_sim_param->set_num_output(1);
    euc_sim_param->set_bias_term(false);
    layer.reset(new CuDNNEuclideanSimilarityLayer<TypeParam>(layer_param));
    layer->blobs().resize(1);
    layer->blobs()[0].reset(new Blob<TypeParam>(1, 3, 1, 3));
    TypeParam* weights_2 = layer->blobs()[0]->mutable_cpu_data();
    for (int c = 0; c < 3; ++c) {
        int i = c * 3;  // 1 x 3 filter
        weights_2[i +  0] = -1;
        weights_2[i +  1] =  0;
        weights_2[i +  2] =  1;
    }
    layer->SetUp(sep_blob_bottom_vec, sep_blob_top_vec);
    layer->Forward(sep_blob_bottom_vec, sep_blob_top_vec);
    // Test equivalence of full and separable filters.
    const TypeParam* top_data = this->blob_top_->cpu_data();
    const TypeParam* sep_top_data = this->blob_top_2_->cpu_data();
    for (int i = 0; i < this->blob_top_->count(); ++i) {
        EXPECT_NEAR(top_data[i], sep_top_data[i], 1e-4);
    }
}

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestGradientCuDNN) {
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    this->blob_bottom_vec_.push_back(this->blob_bottom_2_);
    this->blob_top_vec_.push_back(this->blob_top_2_);
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(2);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    CuDNNEuclideanSimilarityLayer<TypeParam> layer(layer_param);
    GradientChecker<TypeParam> checker(1e-2, 1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

TYPED_TEST(CuDNNEuclideanSimilarityLayerTest, TestGradientGroupCuDNN) {
    LayerParameter layer_param;
    EuclideanSimilarityParameter* euc_sim_param =
        layer_param.mutable_euc_sim_param();
    euc_sim_param->set_kernel_size(3);
    euc_sim_param->set_stride(2);
    euc_sim_param->set_num_output(3);
    euc_sim_param->set_group(3);
    euc_sim_param->mutable_mean_filler()->set_type("gaussian");
    euc_sim_param->mutable_scale_filler()->set_type("gaussian");
    CuDNNEuclideanSimilarityLayer<TypeParam> layer(layer_param);
    GradientChecker<TypeParam> checker(1e-2, 1e-3);
    checker.CheckGradientExhaustive(&layer, this->blob_bottom_vec_,
                                    this->blob_top_vec_);
}

#endif

}  // namespace caffe
